
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <immintrin.h>


#define _POSIX_C_SOURCE 200112L
#include "cputime.h"

#define M 3200
#define N M
#define P M

#define MIN(x, y) ((x) < (y) ? (x) : (y))

float C[M][N];
float A[M][P];
float B[P][N];

static void fill(void)
{
    for (unsigned i = 0; i < M; i++)
        for (unsigned j = 0; j < P; j++)
            A[i][j] = (float) rand();
    for (unsigned i = 0; i < P; i++)
        for (unsigned j = 0; j < N; j++)
            B[i][j] = (float) rand();
}

unsigned int arr[16 * 1024 * 1024];

static unsigned invcache(void)
{
    unsigned int res = 0;
    for (unsigned i = 0; i < sizeof(arr) / sizeof(arr[0]); i++)
        arr[i] = rand();
    for (unsigned i = 0; i < sizeof(arr) / sizeof(arr[0]); i++)
        res += arr[i];
    return res;
}

void matmul(void)
{
    for (unsigned i = 0; i < M; i++)
        for (unsigned j = 0; j < N; j++)
            for (unsigned k = 0; k < P; k++)
                C[i][j] += A[i][k] * B[k][j];
}

void matmul_ichg(void)
{
    for (unsigned i = 0; i < M; i++)
        for (unsigned k = 0; k < P; k++)
            for (unsigned j = 0; j < N; j++)
                C[i][j] += A[i][k] * B[k][j];
}


#define I 64
#define K 32

void matmul_tiled(void)
{
    for (unsigned iT = 0; iT < N; iT += I)
        for (unsigned kT = 0; kT < P; kT += K)
            for (unsigned i = iT; i < MIN(iT + I, N); i++)
                for (unsigned k = kT; k < MIN(kT + K, N); k++) {
                    unsigned j;
                    __m256 a = _mm256_broadcast_ss(&A[i][k]);

                    for (j = 0; j < N; j+=8) {
                        __m256 b, c, d;
                        b = _mm256_load_ps(&B[k][j]);
                        c = _mm256_load_ps(&C[i][j]);
                        c = _mm256_add_ps(c, _mm256_mul_ps(a, b));
                        _mm256_store_ps(&C[i][j], c);
                    }
                    for (; j < N; j++)
                        C[i][j] += A[i][k] * B[k][j];
                }
}

#undef I
#undef K


#define Q 16384
float D[Q][Q];
float E[Q][Q];

#define T  10

void stencil(void)
{
    for (unsigned t = 0; t < T; t++) {
        for (unsigned i = 1; i < Q - 1; i++) {
            for (unsigned j = 1; j < Q - 1; j++) {
                D[i][j] = 0.25f * E[i-1][j  ]
                + 0.25f * E[i  ][j-1]
                + 0.25f * E[i+1][j  ]
                + 0.25f * E[i  ][j+1];
            }
        }
    }
}

#define II 6144
#define JJ 6144

void stencil_tile(void)
{
    for (unsigned t = 0; t < T; t++) {
        for (unsigned ii = 1; ii < (Q - 1) / II; ii++) {
            for (unsigned jj = 1; jj < (Q - 1) / JJ; jj++) {
                for (unsigned i = ii * II; i < (ii + 1) * II; i++) {
                    for (unsigned j = jj * JJ; j < (jj + 1) * JJ; j++) {
                        D[i][j] = 0.25f * ( E[i-1][j  ]
                                + E[i  ][j-1]
                                + E[i+1][j  ]
                                + E[i  ][j+1] );
                    }
                }
            }
        }
    }
}

void stencil_tile_vec(void)
{
    for (unsigned t = 0; t < T; t++) {
        for (unsigned ii = 1; ii < (Q - 1) / II; ii++) {
            for (unsigned jj = 1; jj < (Q - 1) / JJ; jj++) {
                for (unsigned i = ii * II; i < (ii + 1) * II; i++) {
                    static float const c = 0.25;
                    __m256 d = _mm256_broadcast_ss(&c);
                    for (unsigned j = jj * JJ; j < (jj + 1) * JJ; j+=8) {
                        __m256 a, b, c, d;

                        a = _mm256_load_ps(&E[i-1][j]);
                        b = _mm256_loadu_ps(&E[i][j-1]);
                        a = _mm256_add_ps(a, b);
                        c = _mm256_load_ps(&E[i+1][j]);
                        d = _mm256_loadu_ps(&E[i][j+1]);
                        c = _mm256_add_ps(c, d);
                        a = _mm256_add_ps(a, c);
                        a = _mm256_mul_ps(a, d);
                        _mm256_store_ps(&D[i][j], a);
                    }
                }
            }
        }
    }
}



int stencil_exp(void)
{
    for (unsigned i = 0; i < Q; i++)
        for (unsigned j = 0; j < Q; j++)
            E[i][j] = (float) rand();

    int res = 0;
    double start, time;

    res += invcache();
    start = getCPUTime();
    stencil();
    time = getCPUTime() - start;
    printf("normal: %f\n", time);

    res += invcache();
    start = getCPUTime();
    stencil_tile();
    time = getCPUTime() - start;
    printf("tiled: %f\n", time);

    res += invcache();
    start = getCPUTime();
    stencil_tile_vec();
    time = getCPUTime() - start;
    printf("tiled vec: %f\n", time);
    return res;
}

int matmul_exp(void)
{
    double start, time;
    int res = 0;

    fill();

#if 1
    res += invcache();
    start = getCPUTime();
    matmul();
    time = getCPUTime() - start;
    printf("original: %f\n", time);
#endif

    res += invcache();
    start = getCPUTime();
    matmul_ichg();
    time = getCPUTime() - start;
    printf("interchanged: %f\n", time);

    res += invcache();
    start = getCPUTime();
    matmul_tiled();
    time = getCPUTime() - start;
    printf("tiled: %f\n", time);

    return res;
}

int main()
{
    return matmul_exp();
}
